from __future__ import annotations

import dataclasses
import json
import logging
import os
from dataclasses import dataclass

from pkg_resources import resource_stream
from selenium.common.exceptions import (
    ElementNotInteractableException,
    NoSuchElementException,
)
from selenium.webdriver.common.by import By

from . import selenium_helpers as sele
from .buffer import JupyniumBuffer
from .ipynb import cells_to_jupy
from .nvim import NvimInfo
from .rpc_messages import len_pending_messages, receive_message

logger = logging.getLogger(__name__)


update_cell_selection_js_code = (
    resource_stream("jupynium", "js/update_cell_selection.js").read().decode("utf-8")
)

get_cell_inputs_js_code = (
    resource_stream("jupynium", "js/get_cell_inputs.js").read().decode("utf-8")
)


@dataclass
class OnLinesArgs:
    lines: list[str]
    start_row: int
    old_end_row: int
    new_end_row: int

    # Optimisations
    def is_chainable(self, other: "OnLinesArgs") -> bool:
        return (
            self.start_row == other.start_row and self.new_end_row == other.old_end_row
        )

    def chain(self, other: "OnLinesArgs") -> "OnLinesArgs":
        assert self.is_chainable(other)
        return OnLinesArgs(
            other.lines,
            self.start_row,
            self.old_end_row,
            other.new_end_row,
        )


@dataclass
class UpdateSelectionArgs:
    cursor_pos_row: int
    visual_start_row: int


@dataclass
class PrevLazyArgs:
    """
    Args that haven't been processed yet.
    """

    on_lines_args: OnLinesArgs | None = None
    update_selection_args: UpdateSelectionArgs | None = None

    def process(self, nvim_info: NvimInfo, driver, bufnr) -> None:
        if self.on_lines_args is not None:
            process_on_lines_event(nvim_info, driver, bufnr, self.on_lines_args)
            self.on_lines_args = None
        if self.update_selection_args is not None:
            update_cell_selection(nvim_info, driver, bufnr, self.update_selection_args)
            self.update_selection_args = None


@dataclass
class PrevLazyArgsPerBuf:
    """
    Keep separate lazy procedures per buffer.
    """

    data: dict[int, PrevLazyArgs] = dataclasses.field(default_factory=dict)

    def process(self, bufnr: int, nvim_info: NvimInfo, driver) -> None:
        if bufnr in self.data:
            self.data[bufnr].process(nvim_info, driver, bufnr)

    def process_all(self, nvim_info: NvimInfo, driver) -> None:
        for bufnr, lazy_args in self.data.items():
            lazy_args.process(nvim_info, driver, bufnr)
        self.data.clear()

    def lazy_on_lines_event(
        self, nvim_info: NvimInfo, driver, bufnr: int, on_lines_args: OnLinesArgs
    ) -> None:
        if bufnr not in self.data:
            self.data[bufnr] = PrevLazyArgs()

        self.data[bufnr].on_lines_args = lazy_on_lines_event(
            nvim_info, driver, bufnr, self.data[bufnr].on_lines_args, on_lines_args
        )

    def overwrite_update_selection(
        self,
        bufnr: int,
        update_selection_args: UpdateSelectionArgs,
    ) -> None:
        """
        Simply ignore the previous update selection and overwrite to the current one.
        """
        if bufnr not in self.data:
            self.data[bufnr] = PrevLazyArgs()

        self.data[bufnr].update_selection_args = update_selection_args


def process_events(nvim_info: NvimInfo, driver):
    """
    Controls events for a single nvim, and a single cycle of events.

    For example, if there are pending messages do process them all.

    Returns:
        bool status: False if nvim needs to be cleared up.
        rpcrequest event: Notify nvim after cleared up. None if status==True or no need to notify
    """
    # Check if the browser is still alive
    if nvim_info.home_window not in driver.window_handles:
        nvim_info.nvim.lua.Jupynium_notify.error(
            ["Do not close the main page. Detaching the nvim from Jupynium.."],
            async_=True,
        )
        return False, None

    # Check for each buffer
    nvim_info.check_window_alive_and_update(driver)

    # Receive message from this nvim
    # vim.g.jupynium_num_pending_msgs is not reliable
    # but so far it seems to be always smaller than the actual number of pending messages

    prev_lazy_args_per_buf = PrevLazyArgsPerBuf()
    while (
        len_pending_messages(nvim_info.nvim) > 0
        or nvim_info.nvim.vars["jupynium_num_pending_msgs"] > 0
    ):
        event = receive_message(nvim_info.nvim)
        logger.info(f"Event from nvim: {event}")

        if event is None:
            logger.error("Received event=None")
            return False, None

        assert event[1] is not None
        assert event[2] is not None

        if event[0] == "request":
            status, request_event = process_request_event(nvim_info, driver, event)
            if not status:
                return False, request_event
        else:
            # process and update prev_lazy_args
            process_notification_event(nvim_info, driver, event, prev_lazy_args_per_buf)

    # After the loop (here) you need to process the last on_lines event.
    prev_lazy_args_per_buf.process_all(nvim_info, driver)

    return True, None


def start_sync_with_filename(
    bufnr: int,
    filename: str,
    ask: bool,
    content: list[str],
    nvim_info: NvimInfo,
    driver,
):
    """
    Start sync using a filename (not tab index)
    filename has to end with .ipynb
    """
    driver.switch_to.window(nvim_info.home_window)
    sele.wait_until_notebook_list_loaded(driver)

    if filename == "":
        file_found = False
    else:
        notebook_items = driver.find_elements(
            By.CSS_SELECTOR, "#notebook_list > div > div"
        )
        file_found = False
        for notebook_item in notebook_items:
            # is notebook?
            try:
                notebook_item.find_element(By.CSS_SELECTOR, "i.notebook_icon")
            except NoSuchElementException:
                continue

            try:
                notebook_elem = notebook_item.find_element(By.CSS_SELECTOR, "a > span")
            except NoSuchElementException:
                continue
            notebook_name = notebook_elem.text
            if notebook_name == filename:
                prev_windows = set(driver.window_handles)
                driver.execute_script("arguments[0].scrollIntoView();", notebook_elem)
                notebook_elem.click()
                file_found = True
                sele.wait_until_new_window(driver, prev_windows)
                break

    if file_found:
        new_window = set(driver.window_handles) - set(prev_windows)
        assert len(new_window) == 1
        new_window = new_window.pop()

        driver.switch_to.window(new_window)
        sele.wait_until_notebook_loaded(driver)

        if ask:
            sync_input = nvim_info.nvim.eval(
                """input("Press 'v' to sync from n[v]im, 'i' to load from [i]pynb and sync. (v/i/[c]ancel): ")"""
            )
            sync_input = str(sync_input).strip()
        else:
            # if ask == False, sync from vim to ipynb
            sync_input = "v"

        if sync_input in ["v", "V"]:
            # Start sync from vim to ipynb tab
            nvim_info.attach_buffer(bufnr, content, new_window)
            nvim_info.jupbufs[bufnr].full_sync_to_notebook(driver)
        elif sync_input in ["i", "I"]:
            # load from ipynb tab and start sync
            cell_types, texts = driver.execute_script(get_cell_inputs_js_code)
            jupy = cells_to_jupy(cell_types, texts)
            nvim_info.nvim.buffers[bufnr][:] = jupy

            nvim_info.attach_buffer(bufnr, jupy, new_window)
    else:
        new_btn = driver.find_element(By.ID, "new-buttons")
        driver.execute_script("arguments[0].scrollIntoView(true);", new_btn)
        python_btn = driver.find_element(By.ID, "kernel-python3")
        prev_windows = set(driver.window_handles)
        try:
            python_btn.click()
        except ElementNotInteractableException:
            new_btn.click()
            python_btn.click()

        sele.wait_until_new_window(driver, prev_windows)
        new_window = set(driver.window_handles) - prev_windows
        assert len(new_window) == 1
        new_window = new_window.pop()

        driver.switch_to.window(new_window)
        sele.wait_until_notebook_loaded(driver)
        if filename != "":
            driver.execute_script(
                "Jupyter.notebook.rename(arguments[0]);",
                filename,
            )
        # start sync
        nvim_info.attach_buffer(bufnr, content, driver.current_window_handle)
        nvim_info.jupbufs[bufnr].full_sync_to_notebook(driver)


def process_request_event(nvim_info: NvimInfo, driver, event):
    """
    Returns:
        status (bool)
        request_event (rpcrequest event) to notify nvim after cleared up. None if no need to notify
    """
    assert event[0] == "request"
    # Request from nvim
    # send back response

    bufnr = event[2][0]
    event_args = event[2][1:]

    if event[1] == "start_sync":
        filename, ask, content = event_args
        filename = filename.strip()

        filename: str
        if not filename.isnumeric():
            if filename != "" and not filename.lower().endswith(".ipynb"):
                filename += ".ipynb"

            start_sync_with_filename(bufnr, filename, ask, content, nvim_info, driver)
        else:
            # start sync with tab index
            tab_idx = int(filename)
            driver.switch_to.window(driver.window_handles[tab_idx - 1])

            continue_input = "y"
            if ask:
                continue_input = nvim_info.nvim.eval(
                    "input('This will remove all content from the Notebook. Continue? (y/n): ')"
                )
            if continue_input in ["y", "Y"]:
                nvim_info.attach_buffer(bufnr, content, driver.current_window_handle)
                nvim_info.jupbufs[bufnr].full_sync_to_notebook(driver)
            else:
                event[3].send("N")
                event[3] = None
    elif event[1] == "load_from_ipynb_tab":
        (tab_idx,) = event_args
        if tab_idx > len(driver.window_handles) or tab_idx < 1:
            nvim_info.nvim.lua.Jupynium_notify.error(
                [f"Tab {tab_idx} doesn't exist."],
                async_=True,
            )
            event[3].send("N")
            return False, None
        driver.switch_to.window(driver.window_handles[tab_idx - 1])

        cell_types, texts = driver.execute_script(get_cell_inputs_js_code)
        jupy = cells_to_jupy(cell_types, texts)
        nvim_info.nvim.buffers[bufnr][:] = jupy
        logger.info(f"Loaded ipynb to the nvim buffer.")

    elif event[1] == "VimLeavePre":
        logger.info("Nvim closed. Clearing nvim")
        return False, event[3]

    if event[3] is not None:
        event[3].send("OK")

    return True, None


def skip_bloated(nvim_info: NvimInfo):
    if nvim_info.nvim.vars["jupynium_message_bloated"]:
        logger.info("Message bloated. Skipping..")
        if len_pending_messages(nvim_info.nvim) == 0:
            nvim_info.nvim.vars["jupynium_num_pending_msgs"] = 0
            nvim_info.nvim.vars["jupynium_message_bloated"] = False

            logger.info("Reloading all buffers from nvim")
            for buf_id in nvim_info.jupbufs.keys():
                nvim_info.nvim.lua.Jupynium_grab_entire_buffer(buf_id)

        return True
    return False


def lazy_on_lines_event(
    nvim_info: NvimInfo,
    driver,
    bufnr: int,
    previous_on_lines: OnLinesArgs | None,
    current_on_lines: OnLinesArgs,
):
    """
    Lazy-process on_lines events.
    ----
    Often, completion plugins like coc.nvim and nvim-cmp spams on_lines events.
    But they will have the same (bufnr, start_row, old_end_row, new_end_row) values.
    If the series of line changes are chainable, we can just process the last one.
    In order to do that, process the previous on_lines only if current on_lines is not chainable.
    After the loop you need to process the last one as well.
    """
    if previous_on_lines is None:
        return current_on_lines

    if previous_on_lines.is_chainable(current_on_lines):
        return previous_on_lines.chain(current_on_lines)

    process_on_lines_event(nvim_info, driver, bufnr, previous_on_lines)
    return current_on_lines


def process_on_lines_event(
    nvim_info: NvimInfo, driver, bufnr, on_lines_args: OnLinesArgs
):
    driver.switch_to.window(nvim_info.window_handles[bufnr])

    nvim_info.jupbufs[bufnr].process_on_lines(
        driver,
        True,
        on_lines_args.lines,
        on_lines_args.start_row,
        on_lines_args.old_end_row,
        on_lines_args.new_end_row,
    )


# flake8: noqa: C901
def process_notification_event(
    nvim_info: NvimInfo,
    driver,
    event,
    prev_lazy_args_per_buf: PrevLazyArgsPerBuf | None = None,
):
    assert event[0] == "notification"

    if skip_bloated(nvim_info):
        return

    bufnr = event[2][0]
    event_args = event[2][1:]
    if event[1] == "on_lines":
        current_on_lines = OnLinesArgs(*event_args)

        if prev_lazy_args_per_buf is None:
            process_on_lines_event(nvim_info, driver, bufnr, current_on_lines)
        else:
            prev_lazy_args_per_buf.lazy_on_lines_event(
                nvim_info, driver, bufnr, current_on_lines
            )
    elif event[1] in [
        "CursorMoved",
        "CursorMovedI",
        "visual_enter",
        "visual_leave",
    ]:
        current_args = UpdateSelectionArgs(*event_args)

        if prev_lazy_args_per_buf is not None:
            # Lazy. Process at the end of event loop cycle or if needed for the next event.
            prev_lazy_args_per_buf.overwrite_update_selection(bufnr, current_args)
        else:
            update_cell_selection(nvim_info, driver, bufnr, current_args)
    else:
        # For all the other events, it requires lazy events to be performed in advance.
        if prev_lazy_args_per_buf is not None:
            prev_lazy_args_per_buf.process(bufnr, nvim_info, driver)

        if event[1] == "scroll_ipynb":
            (scroll,) = event_args
            driver.switch_to.window(nvim_info.window_handles[bufnr])

            driver.execute_script(
                "Jupyter.notebook.scroll_manager.animation_speed = 0; Jupyter.notebook.scroll_manager.scroll_some(arguments[0]);",
                scroll,
            )
        elif event[1] == "save_ipynb":
            driver.switch_to.window(nvim_info.window_handles[bufnr])

            driver.execute_script("Jupyter.notebook.save_notebook();")
            driver.execute_script("Jupyter.notebook.save_checkpoint();")
        elif event[1] == "BufWritePre":
            (buf_filepath,) = event_args
            driver.switch_to.window(nvim_info.window_handles[bufnr])

            driver.execute_script("Jupyter.notebook.save_notebook();")
            driver.execute_script("Jupyter.notebook.save_checkpoint();")

            if (
                ".ju." in buf_filepath
                and nvim_info.nvim.vars["jupynium_auto_download_ipynb"]
            ):
                output_ipynb_path = os.path.splitext(buf_filepath)[0]
                output_ipynb_path = os.path.splitext(output_ipynb_path)[0]
                output_ipynb_path += ".ipynb"

                download_ipynb(driver, nvim_info, bufnr, output_ipynb_path)
        elif event[1] == "download_ipynb":
            (buf_filepath, filename) = event_args
            assert buf_filepath != ""

            if filename is not None and filename != "":
                if os.path.isabs(filename):
                    output_ipynb_path = filename
                else:
                    output_ipynb_path = os.path.join(
                        os.path.dirname(buf_filepath), filename
                    )

                if not output_ipynb_path.endswith(".ipynb"):
                    output_ipynb_path += ".ipynb"
            else:
                output_ipynb_path = os.path.splitext(buf_filepath)[0]
                output_ipynb_path = os.path.splitext(output_ipynb_path)[0]
                output_ipynb_path += ".ipynb"

            download_ipynb(driver, nvim_info, bufnr, output_ipynb_path)

        elif event[1] == "toggle_selected_cells_outputs_scroll":
            driver.switch_to.window(nvim_info.window_handles[bufnr])
            driver.execute_script(
                "Jupyter.notebook.toggle_cells_outputs_scroll(Jupyter.notebook.get_selected_cells_indices())"
            )
        elif event[1] == "execute_selected_cells":
            driver.switch_to.window(nvim_info.window_handles[bufnr])
            driver.execute_script("Jupyter.notebook.execute_selected_cells();")
        elif event[1] == "clear_selected_cells_outputs":
            driver.switch_to.window(nvim_info.window_handles[bufnr])
            driver.execute_script(
                "Jupyter.notebook.clear_cells_outputs(Jupyter.notebook.get_selected_cells_indices())"
            )
            # driver.execute_script("Jupyter.notebook.clear_output();")
        elif event[1] == "scroll_to_cell":
            (cursor_pos_row,) = event_args
            scroll_to_cell(driver, nvim_info, bufnr, cursor_pos_row)

        elif event[1] == "grab_entire_buf":
            # Refresh entire buffer from nvim
            # But do not necessarily sync to Jupyter Notebook
            # because it happens when you spam events and it slows down.
            (content,) = event_args

            nvim_info.jupbufs[bufnr] = JupyniumBuffer(content)
            if driver.current_window_handle == nvim_info.window_handles[bufnr]:
                nvim_info.jupbufs[bufnr].full_sync_to_notebook(driver)

        elif event[1] == "BufWinLeave":
            logger.info("Buffer closed on nvim. Closing on Jupyter Notebook")
            # driver.switch_to.window(nvim_info.window_handles[buf])
            # driver.close()
            nvim_info.detach_buffer(bufnr, driver)

        elif event[1] == "stop_sync":
            logger.info(f"Received stop_sync request: bufnr = {bufnr}")
            nvim_info.detach_buffer(bufnr, driver)


def update_cell_selection(
    nvim_info: NvimInfo, driver, bufnr, update_selection_args: UpdateSelectionArgs
):

    cursor_pos_row, visual_start_row = dataclasses.astuple(update_selection_args)

    if nvim_info.jupbufs[bufnr].num_cells == 1:
        # No autoscroll for markdown mode
        driver.switch_to.window(nvim_info.window_handles[bufnr])
    else:
        # Which cell?
        try:
            cell_index, _, _ = nvim_info.jupbufs[bufnr].get_cell_index_from_row(
                cursor_pos_row
            )
        except IndexError:
            if cursor_pos_row == nvim_info.jupbufs[bufnr].num_rows:
                # Nvim 0.8.1 has a bug where it
                # doesn't send on_byte event but it
                # sends CursorMoved event when
                # you create a new line with 'o' at the end of file.
                # So we have to manually add a new line
                nvim_info.jupbufs[bufnr].buf.append("")
                nvim_info.jupbufs[bufnr].process_on_lines(
                    driver,
                    True,
                    [""],
                    nvim_info.jupbufs[bufnr].num_rows,
                    nvim_info.jupbufs[bufnr].num_rows,
                    nvim_info.jupbufs[bufnr].num_rows + 1,
                )
                cell_index, _, _ = nvim_info.jupbufs[bufnr].get_cell_index_from_row(
                    cursor_pos_row
                )
            else:
                logger.exception("Cannot find cell index given row")
                return

        cell_index_visual, _, _ = nvim_info.jupbufs[bufnr].get_cell_index_from_row(
            visual_start_row
        )

        # select the cell
        # 0-th cell is not a cell, but it's the header. You can put anything above and it won't be synced to Jupyter Notebook.
        cell_index = max(cell_index - 1, 0)
        cell_index_visual = max(cell_index_visual - 1, 0)

        driver.switch_to.window(nvim_info.window_handles[bufnr])

        selection_updated = driver.execute_script(
            update_cell_selection_js_code, cell_index, cell_index_visual
        )

        if selection_updated:
            autoscroll_enable = nvim_info.nvim.vars.get(
                "jupynium_autoscroll_enable", True
            )
            if autoscroll_enable:
                autoscroll_mode = nvim_info.nvim.vars.get(
                    "jupynium_autoscroll_mode", "always"
                )

                if autoscroll_mode == "always":
                    do_scroll = True
                else:
                    do_scroll = not driver.execute_script(
                        "return Jupyter.notebook.scroll_manager.is_cell_visible(Jupyter.notebook.get_cell(arguments[0]));",
                        cell_index,
                    )

                if do_scroll:
                    # scroll to cell
                    top_margin_percent = nvim_info.nvim.vars.get(
                        "jupynium_autoscroll_cell_top_margin_percent", 0
                    )
                    driver.execute_script(
                        "Jupyter.notebook.scroll_cell_percent(arguments[0], arguments[1], 0);",
                        cell_index,
                        top_margin_percent,
                    )


def download_ipynb(driver, nvim_info, bufnr, output_ipynb_path):
    driver.switch_to.window(nvim_info.window_handles[bufnr])

    with open(output_ipynb_path, "w") as f:
        json.dump(
            driver.execute_script(
                "return Jupyter.notebook.toJSON();",
            ),
            f,
            indent=4,
        )
        nvim_info.nvim.lua.Jupynium_notify.info(
            ["Downloaded ipynb file to", output_ipynb_path],
            async_=True,
        )
        logger.info(f"Downloaded ipynb to {output_ipynb_path}")


def scroll_to_cell(driver, nvim_info, bufnr, cursor_pos_row):
    # Which cell?
    cell_index, _, _ = nvim_info.jupbufs[bufnr].get_cell_index_from_row(cursor_pos_row)

    cell_index = max(cell_index - 1, 0)

    driver.switch_to.window(nvim_info.window_handles[bufnr])

    top_margin_percent = nvim_info.nvim.vars.get(
        "jupynium_autoscroll_cell_top_margin_percent", 0
    )
    driver.execute_script(
        "Jupyter.notebook.scroll_cell_percent(arguments[0], arguments[1], 0);",
        cell_index,
        top_margin_percent,
    )



from __future__ import annotations

import json


def load_ipynb(ipynb_path):
    with open(ipynb_path, "r") as f:
        ipynb = json.load(f)
    return ipynb


def read_ipynb_texts(ipynb):
    texts = []
    cell_types = []
    for cell in ipynb["cells"]:
        cell_types.append(cell["cell_type"])
        texts.append("".join(cell["source"]))
    return cell_types, texts


def cells_to_jupy(cell_types, texts):
    cell_types_previous = ["code"] + cell_types[:-1]

    jupy: list[str] = []

    for cell_type_previous, cell_type, text in zip(
        cell_types_previous, cell_types, texts
    ):
        if cell_type == "code":
            if cell_type_previous == "code":
                jupy.append("# %%")
            else:
                jupy.append('%%"""')
        else:
            if cell_type_previous == "code":
                jupy.append('"""%%')
            else:
                jupy.append("# %%%")

        jupy.extend(text.split("\n"))

    return jupy


def ipynb2jupy(ipynb):
    cell_types, texts = read_ipynb_texts(ipynb)
    return cells_to_jupy(cell_types, texts)



#!/use/bin/env python3
from __future__ import annotations

import argparse
import configparser
import logging
import os
import secrets
import subprocess
import sys
import tempfile
import time
import traceback
from pathlib import Path
from urllib.parse import urlparse

import coloredlogs
import git
import persistqueue
import verboselogs
from git.exc import InvalidGitRepositoryError
from persistqueue.exceptions import Empty
from selenium import webdriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from .. import __version__
from .. import selenium_helpers as sele
from ..definitions import persist_queue_path
from ..events_control import process_events
from ..nvim import NvimInfo
from ..process import already_running_pid
from ..pynvim_helpers import attach_and_init

logger = verboselogs.VerboseLogger(__name__)

SOURCE_DIR = Path(os.path.dirname(os.path.abspath(__file__)))


def webdriver_firefox(
    profiles_ini_path="~/.mozilla/firefox/profiles.ini", profile_name=None
):
    """
    profiles.ini path is used to remember the last session (password, etc.)
    Args:
        profiles_ini_path: Path to profiles.ini
        profile_name: Profile name in profiles.ini. If None, use the default profile.
    """
    # Read firefox profile path from profiles.ini
    profile_path = None

    if profiles_ini_path is not None:
        profiles_ini_path = Path(profiles_ini_path).expanduser()
        if profiles_ini_path.exists():
            config = configparser.ConfigParser()
            config.read(profiles_ini_path)
            for section in config.sections():
                if section.startswith("Profile"):
                    try:
                        if profile_name is None or profile_name == "":
                            if config[section]["Default"] == "1":
                                section_ = section
                                break
                        else:
                            if config[section]["Name"] == profile_name:
                                section_ = section
                                break
                    except KeyError:
                        pass
            else:
                section_ = None
                profile_path = None

            if section_ is not None:
                try:
                    profile_path = config[section_]["Path"]
                    is_relative = config[section_]["IsRelative"]
                    if is_relative == "1":
                        profile_path = profiles_ini_path.parent / profile_path
                    else:
                        profile_path = Path(profile_path)
                    if not profile_path.exists():
                        profile_path = None
                except KeyError:
                    profile_path = None

    logger.info(f"Using firefox profile: {profile_path}")

    profile = webdriver.FirefoxProfile(profile_path)
    profile.set_preference("browser.link.open_newwindow", 3)
    profile.set_preference("browser.link.open_newwindow.restriction", 0)
    # profile.setAlwaysLoadNoFocusLib(True);
    return webdriver.Firefox(profile, service_log_path=os.path.devnull)


# def webdriver_safari():
#     return webdriver.Safari()
#
#
# def webdriver_chrome():
#     from selenium.webdriver.chrome.service import Service
#     from webdriver_manager.chrome import ChromeDriverManager
#
#     options = webdriver.ChromeOptions()
#     return webdriver.Chrome(
#         service=Service(ChromeDriverManager().install()), options=options
#     )


def get_parser():
    parser = argparse.ArgumentParser(
        description="Control Jupyter Notebook with Neovim",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--notebook_URL",
        default="localhost:8888",
        help="Jupyter Notebook URL to open from the Selenium browser",
    )
    parser.add_argument(
        "--nvim_listen_addr",
        # default="localhost:18898",
        help="TCP or socket path (file path)",
    )
    parser.add_argument(
        "--attach_only",
        action="store_true",
        help="Attach to an existing Jupynium instance."
        "If False, start a new instance or attach to an existing one."
        "If True, all arguments except nvim_listen_addr and notebook_URL are ignored.",
    )
    parser.add_argument(
        "--sleep_time_idle",
        type=float,
        default=0.1,
        help="Sleep time when there is no event to process.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Print version and exit.",
    )
    parser.add_argument(
        "--check_running",
        action="store_true",
        help="Print pid if Jupynium is running. Otherwise, print nothing.",
    )
    parser.add_argument(
        "--firefox_profiles_ini_path",
        default="~/.mozilla/firefox/profiles.ini",
        help="Path to firefox profiles.ini which will be used to remember the last session (password, etc.)",
    )
    parser.add_argument(
        "--firefox_profile_name",
        help="Firefox profile name. If None, use the default profile.",
    )
    parser.add_argument(
        "--jupyter_command",
        type=str,
        nargs="+",
        default=["jupyter"],
        help="Command to start Jupyter Notebook (but without notebook)."
        "To use conda env, use `--jupyter_command ~/miniconda3/envs/env_name/bin/jupyter`."
        "Don't use `conda run ..` as it won't be killed afterwards (it opens another process with different pid so it's hard to keep track of it.)"
        "It is used only when the --notebook_URL is localhost, and is not running.",
    )
    parser.add_argument(
        "--notebook_dir",
        type=str,
        help="When jupyter notebook has started using --jupyter_command, the root dir will be this."
        "If None, open at a git dir of nvim's buffer path and still navigate to the buffer dir."
        "(e.g. localhost:8888/tree/path/to/buffer)",
    )

    # parser.add_argument(
    #     "--browser",
    #     choices=["firefox", "safari", "chrome"],
    #     default="firefox",
    #     help="Browser to use. Best with firefox. Chrome may often steal the focus.",
    # )
    return parser


def start_if_running_else_clear(args, q: persistqueue.UniqueQ):
    # If Jupynium is already running, send args and quit.
    if already_running_pid():
        if args.nvim_listen_addr is not None:
            q.put(args)
            logger.info(
                "Jupynium is already running. Attaching to the running process."
            )
            return 0
        else:
            logger.info(
                "Jupynium is already running. Attach nvim using --nvim_listen_addr"
            )
            return 1
    else:
        if args.attach_only:
            logger.error(
                "Jupynium is not running. Remove --attach_only option to start a new instance."
            )
            return 1

    # If Jupynium is not running, clear the message queue before starting.
    while True:
        try:
            _ = q.get(block=False)
        except Empty:
            break

    return None


def attach_new_neovim(
    driver,
    new_args,
    nvims: dict[str, NvimInfo],
    URL_to_home_windows: dict[str, str],
):
    logger.info(f"New nvim wants to attach: {new_args}")
    if new_args.nvim_listen_addr in nvims:
        logger.info("Already attached.")
    else:
        try:
            nvim = attach_and_init(new_args.nvim_listen_addr)
            if new_args.notebook_URL in URL_to_home_windows.keys():
                home_window = URL_to_home_windows[new_args.notebook_URL]
            else:
                prev_num_windows = len(driver.window_handles)
                driver.switch_to.new_window("tab")
                driver.get(new_args.notebook_URL)

                # Wait for the notebook to load
                driver_wait = WebDriverWait(driver, 10)
                driver_wait.until(EC.number_of_windows_to_be(prev_num_windows + 1))
                sele.wait_until_loaded(driver)

                home_window = driver.current_window_handle
                URL_to_home_windows[new_args.notebook_URL] = home_window

            nvim_info = NvimInfo(nvim, home_window)
            nvims[new_args.nvim_listen_addr] = nvim_info
        except Exception:
            logger.exception("Exception occurred while attaching a new nvim. Ignoring.")


def nvims_teardown(nvims):
    # Before exiting, tell vim about it.
    # Otherwise vim will need to communicate once more to find out.
    try:
        for nvim in nvims.values():
            nvim.nvim.lua.Jupynium_reset_channel(async_=True)
    except Exception:
        # Even if you fail it's not a big problem
        pass


def generate_notebook_token():
    return secrets.token_urlsafe(16)


def exception_no_notebook(notebook_URL, nvim):
    logger.exception(
        f"Exception occurred. Are you sure you're running Jupyter Notebook at {notebook_URL}? Use --jupyter_command to specify the command to start Jupyter Notebook."
    )
    if nvim is not None:
        nvim.lua.Jupynium_notify.error(
            [
                "Can't connect to Jupyter Notebook.",
                f"Are you sure you're running Jupyter Notebook at {notebook_URL}?",
                "Use jupyter_command to specify the command to start Jupyter Notebook.",
            ],
        )
        nvim.lua.Jupynium_reset_channel()

    sys.exit(1)


def kill_notebook_proc(notebook_proc):
    """
    Kill the notebook process.
    Used if we opened a Jupyter Notebook server using the --jupyter_command and when no server is running.
    """
    if notebook_proc is not None:
        notebook_proc.terminate()
        # notebook_proc.kill()
        notebook_proc.wait()
        logger.info("Jupyter Notebook server has been killed.")


def fallback_open_notebook_server(
    notebook_port, jupyter_command, notebook_dir, nvim, driver
):
    # Fallback: if the URL is localhost and if selenium can't connect,
    # open the Jupyter Notebook server and even start syncing.
    rel_dir = ""

    if notebook_dir is None or notebook_dir == "":
        notebook_dir = None
        if nvim is not None:
            # Root dir of the notebook is either the buffer's dir or the git dir.
            buffer_path = str(nvim.eval("expand('%:p')"))
            buffer_dir = os.path.dirname(buffer_path)
            try:
                repo = git.Repo(buffer_dir, search_parent_directories=True)
                notebook_dir = repo.working_tree_dir
                rel_dir = os.path.relpath(buffer_dir, notebook_dir)
            except InvalidGitRepositoryError:
                notebook_dir = buffer_dir

    notebook_token = generate_notebook_token()
    notebook_args = [
        "notebook",
        "--port",
        str(notebook_port),
        "--no-browser",
        f"--NotebookApp.token",
        notebook_token,
    ]

    if notebook_dir is not None:
        # notebook_args += [f"--ServerApp.root_dir={root_dir}"]
        notebook_args += ["--NotebookApp.notebook_dir", notebook_dir]

    try:
        # strip commands because we need to escape args with dashes.
        # e.g. --jupyter_command conda run ' --no-capture-output' ' -n' env_name jupyter
        # However, conda run will run process with another pid so it won't work well here. Don't use it.

        jupyter_command = [command.strip() for command in jupyter_command]
        jupyter_command[0] = os.path.expanduser(jupyter_command[0])

        jupyter_stdout = tempfile.NamedTemporaryFile()
        logger.info(f"Writing Jupyter Notebook server log to: {jupyter_stdout.name}")
        notebook_proc = subprocess.Popen(
            jupyter_command + notebook_args, stdout=jupyter_stdout
        )
    except FileNotFoundError:
        # Command doesn't exist
        exception_no_notebook(f"localhost:{notebook_port}", nvim)

    time.sleep(1)
    for _ in range(20):
        try:
            driver.get(
                f"localhost:{notebook_port}/tree/{rel_dir}?token={notebook_token}"
            )
            break
        except WebDriverException:
            poll = notebook_proc.poll()
            if poll is not None:
                # Process finished
                exception_no_notebook(f"localhost:{notebook_port}", nvim)

        time.sleep(0.3)
    else:
        # Process still running but timeout for connecting to notebook. Maybe wrong command?
        kill_notebook_proc(notebook_proc)
        exception_no_notebook(f"localhost:{notebook_port}", nvim)
    return notebook_proc


# flake8: noqa: C901
def main():
    coloredlogs.install(
        fmt="%(name)s: %(lineno)4d - %(levelname)s - %(message)s", level=logging.INFO
    )

    parser = get_parser()
    args = parser.parse_args()

    if args.version:
        print(f"Jupynium v{__version__}")
        sys.exit(0)

    if args.check_running:
        pid = already_running_pid()
        if pid > 0:
            print(pid)
        sys.exit(0)

    q = persistqueue.UniqueQ(persist_queue_path)
    return_code = start_if_running_else_clear(args, q)
    if return_code is not None:
        sys.exit(return_code)

    nvim = None
    if args.nvim_listen_addr is not None:
        try:
            nvim = attach_and_init(args.nvim_listen_addr)
        except Exception:
            logger.exception("Exception occurred")
            sys.exit(1)

    nvims = {}
    notebook_proc = None
    try:
        # Open selenium
        # If you load with Chrome, it will annoyingly set focus to the browser
        # when you open or change tab (basically happens every time you type)

        # If you load with Safari, it won't let you interact with the browser.

        with webdriver_firefox(
            args.firefox_profiles_ini_path, args.firefox_profile_name
        ) as driver:
            try:
                driver.get(args.notebook_URL)
            except WebDriverException:
                notebook_URL = args.notebook_URL
                if "://" not in args.notebook_URL:
                    notebook_URL = "http://" + notebook_URL
                url = urlparse(notebook_URL)
                if url.port is not None and url.hostname in ["localhost", "127.0.0.1"]:
                    notebook_proc = fallback_open_notebook_server(
                        url.port, args.jupyter_command, args.notebook_dir, nvim, driver
                    )

                else:
                    # Not localhost, so not trying to start the notebook server.
                    exception_no_notebook(args.notebook_URL, nvim)

            # Wait for the notebook to load
            driver_wait = WebDriverWait(driver, 10)
            driver_wait.until(EC.number_of_windows_to_be(1))
            sele.wait_until_loaded(driver)

            home_window = driver.current_window_handle

            URL_to_home_windows = {args.notebook_URL: home_window}
            if args.nvim_listen_addr is not None and nvim is not None:
                nvims = {args.nvim_listen_addr: NvimInfo(nvim, home_window)}
            else:
                logger.info(
                    "No nvim attached. Waiting for nvim to attach. Run jupynium --nvim_listen_addr /tmp/example (use `:echo v:servername` of nvim)"
                )

            while not sele.is_browser_disconnected(driver):
                try:
                    del_list = []
                    for nvim_listen_addr, nvim_info in nvims.items():
                        try:
                            status, rpcrequest_event = process_events(nvim_info, driver)
                        except OSError:
                            logger.info("Nvim has been closed. Detaching nvim.")
                            del_list.append((nvim_listen_addr, None))
                        except Exception:
                            logger.exception(
                                "Uncaught exception occurred while processing events. Detaching nvim."
                            )
                            del_list.append((nvim_listen_addr, None))

                        else:
                            if not status:
                                del_list.append((nvim_listen_addr, rpcrequest_event))

                    # Remove nvim if browser is closed
                    # Must do it outside the loop
                    for listen_addr, rpcrequest_event in del_list:
                        nvims[listen_addr].close(driver)
                        if rpcrequest_event is not None:
                            rpcrequest_event.send("OK")
                        del nvims[listen_addr]

                    # Check if a new newvim instance wants to attach to this server.
                    try:
                        new_args = q.get(block=False)
                    except Empty:
                        pass
                    else:
                        attach_new_neovim(driver, new_args, nvims, URL_to_home_windows)

                    time.sleep(args.sleep_time_idle)
                except WebDriverException:
                    break

            logger.info("Browser disconnected. Quitting Jupynium.")

    except Exception:
        logger.exception("Exception occurred")
        for nvim in nvims.values():
            nvim.nvim.lua.Jupynium_notify.error(
                [
                    "Closed due to exception:",
                    None,
                    "```",
                    traceback.format_exc(),
                    "```",
                ],
                async_=True,
            )
    else:
        logger.success("Piecefully closed as the browser is closed.")

    nvims_teardown(nvims)
    kill_notebook_proc(notebook_proc)


if __name__ == "__main__":
    main()



import os
import subprocess
import tempfile

import pytest

from jupynium import pynvim_helpers
from jupynium.buffer import JupyniumBuffer


@pytest.fixture(scope="session")
def jupbuf1():
    return JupyniumBuffer(["a", "b", "c", "'''%%%", "d", "%%'''", "f"])


@pytest.fixture(scope="session")
def nvim_1():
    with tempfile.TemporaryDirectory() as tmp:
        path = os.path.join(tmp, "nvim")
        nvim_proc = subprocess.Popen(
            ["nvim", "--clean", "--headless", "--listen", path]
        )
        # os.system(f"nvim --clean --headless --listen {path} &")
        nvim = pynvim_helpers.attach_and_init(path)

        yield nvim

        # Teardown

        # nvim.quit() sometimes hangs
        nvim_proc.kill()



from __future__ import annotations

import logging
from dataclasses import dataclass, field

import pynvim

from .buffer import JupyniumBuffer

logger = logging.getLogger(__name__)


@dataclass
class NvimInfo:
    """Information about the pynvim and Jupynium instance."""

    nvim: pynvim.Nvim
    home_window: str
    jupbufs: dict[int, JupyniumBuffer] = field(default_factory=dict)  # key = buffer ID
    window_handles: dict[int, str] = field(default_factory=dict)  # key = buffer ID

    def attach_buffer(self, buf_id, content: list[str], window_handle):
        if buf_id in self.jupbufs or buf_id in self.window_handles:
            logger.warning(f"Buffer {buf_id} is already attached")

        self.jupbufs[buf_id] = JupyniumBuffer(content)
        self.window_handles[buf_id] = window_handle

    def detach_buffer(self, buf_id, driver):
        if buf_id in self.jupbufs:
            del self.jupbufs[buf_id]
        if buf_id in self.window_handles:
            if self.window_handles[buf_id] in driver.window_handles:
                driver.switch_to.window(self.window_handles[buf_id])
                driver.close()
                driver.switch_to.window(self.home_window)
            del self.window_handles[buf_id]

    def check_window_alive_and_update(self, driver):
        detach_buffer_list = []
        for buf_id, window in self.window_handles.items():
            if window not in driver.window_handles:
                self.nvim.lua.Jupynium_notify.error(
                    [
                        "Notebook closed.",
                        f"Detaching the buffer {buf_id} from Jupynium..",
                    ],
                    async_=True,
                )
                self.nvim.lua.Jupynium_stop_sync(buf_id)
                detach_buffer_list.append(buf_id)

        for buf_id in detach_buffer_list:
            self.detach_buffer(buf_id, driver)

    def close(self, driver):
        try:
            self.nvim.lua.Jupynium_reset_channel(async_=True)
        except Exception:
            # Even if you fail it's not a big problem
            pass

        for buf_id in list(self.jupbufs.keys()):
            self.detach_buffer(buf_id, driver)

    def __len__(self):
        return len(self.jupbufs)



import logging

from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

logger = logging.getLogger(__name__)


def wait_until_notebook_loaded(driver, timeout=10):
    """Wait until the Jupyter Notebook is loaded."""
    try:
        WebDriverWait(driver, timeout).until(
            EC.presence_of_element_located((By.ID, "notebook-container"))
        )
    except TimeoutException:
        logger.exception("Timed out waiting for page to load")
        driver.quit()


def wait_until_notebook_list_loaded(driver, timeout=10):
    """Wait until the Jupyter Notebook home page (list of files) is loaded."""
    try:
        WebDriverWait(driver, timeout).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "#notebook_list > div > div > a > span")
            )
        )
    except TimeoutException:
        logger.exception("Timed out waiting for page to load")
        driver.quit()


def wait_until_loaded(driver, timeout=10):
    """Wait until the page is ready."""
    try:
        WebDriverWait(driver, timeout).until(
            lambda d: d.execute_script("return document.readyState") == "complete"
        )
    except TimeoutException:
        logger.exception("Timed out waiting for page to load")
        driver.quit()


def wait_until_new_window(driver, current_handles, timeout=10):
    """Wait until the page is ready."""
    try:
        WebDriverWait(driver, timeout).until(EC.new_window_is_opened(current_handles))
    except TimeoutException:
        logger.exception("Timed out waiting for a new window to open")
        driver.quit()


def is_browser_disconnected(driver):
    """Check if the browser is disconnected."""
    # get_log is not supported by Firefox
    # return driver.get_log("driver")[-1]["message"] == DISCONNECTED_MSG    # not supported by Firefox
    try:
        _ = driver.window_handles
        return False
    except Exception:
        return True



import time

from pynvim import Nvim

from jupynium.rpc_messages import receive_all_pending_messages, receive_message


def test_event_default_variables(nvim_1: Nvim):
    assert nvim_1.vars["jupynium_channel_id"] > 0
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 0


def test_event_before_start_sync(nvim_1: Nvim):
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    nvim_1.feedkeys("i# %%\nabc")
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 0


def test_event_start_sync_cancel(nvim_1: Nvim):
    nvim_1.lua.Jupynium_start_sync(async_=True)
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 1
    event = receive_message(nvim_1)
    assert event is not None
    assert event[0] == "request"
    assert event[1] == "start_sync"

    event[3].send("N")  # Not OK is cancel
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    nvim_1.feedkeys("i# %%\nabc")
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 0


def test_event_start_sync(nvim_1: Nvim):
    nvim_1.lua.Jupynium_start_sync(async_=True)
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 1
    event = receive_message(nvim_1)
    assert event is not None
    assert event[0] == "request"
    assert event[1] == "start_sync"

    event[3].send("OK")  # Not OK is cancel
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    nvim_1.feedkeys("i# %%\nabc")
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))

    assert nvim_1.vars["jupynium_num_pending_msgs"] > 0

    count_cursormoved_i = 0

    time.sleep(0.5)
    events = receive_all_pending_messages(nvim_1)
    for event in events:
        if event[0] == "notification" and event[1] == "CursorMovedI":
            count_cursormoved_i += 1

    assert (
        count_cursormoved_i > 0
    )  # CursorMovedI should be triggered. I think it gets triggered only once after the feedkeys call


def test_event_stop_sync(nvim_1: Nvim):
    nvim_1.lua.Jupynium_stop_sync(async_=True)
    assert nvim_1.vars["jupynium_num_pending_msgs"] == 1
    event = receive_message(nvim_1)
    assert event is not None
    assert event[0] == "notification"
    assert event[1] == "stop_sync"

    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))
    nvim_1.feedkeys("i# %%\nabc")
    nvim_1.feedkeys(nvim_1.replace_termcodes("<esc>", True, True, True))

    assert nvim_1.vars["jupynium_num_pending_msgs"] == 0



import logging

from pynvim import Nvim

logger = logging.getLogger(__name__)


def len_pending_messages(nvim: Nvim):
    return len(nvim._session._pending_messages)


def receive_message(nvim: Nvim):
    event = nvim.next_message()

    # This value is not strictly accurate.
    # It is correct most of the time, but when you send too many messages at once (e.g. change visual selection using mouse) it can be wrong (smaller than the actual number)
    # It is only here to detect bloated state (too many messages)
    nvim.vars["jupynium_num_pending_msgs"] = max(
        nvim.vars.get("jupynium_num_pending_msgs", 0) - 1, 0
    )

    return event


def receive_all_pending_messages(nvim: Nvim):
    """
    It doesn't guarantee to grab all messages that are previously sent. Maybe the last one or two may still be in process.
    """
    events = []
    while len_pending_messages(nvim) > 0:
        events.append(receive_message(nvim))

    return events


"""
def event_to_dict(event):
    event_dict = {}
    if event is None:
        logger.error("event is None")
        return event_dict

    event_dict["type"] = event[0]  # request / notification
    if event[0] == "request":
        event_dict["request"] = event[3]

    # on_bytes / on_bytes_remove, CursorMoved, CursorMovedI, visual_enter, visual_leave, grab_entire_buf, VimLeave
    event_dict["name"] = event[1]
    event_args = {}

    if event[1] == "start_sync":
        (
            _,
            event_args["filename"],
            event_args["content"],
        ) = event[2]
    elif event[1] in [
        "stop_sync",
        "save_ipynb",
        "toggle_selected_cells_output_scroll",
        "execute_selected_cells",
        "clear_selected_cells_outputs",
        "VimLeavePre",
        "BufWinLeave",
    ]:
        pass
    elif event[1] == "on_lines":
        (
            _,
            event_args["lines"],
            event_args["start_row"],
            event_args["old_end_row"],
            event_args["new_end_row"],
            event_args["old_byte_size"],
        ) = event[2]
    elif event[1] == "scroll_ipynb":
        (
            _,
            event_args["scroll"],
        ) = event[2]
    elif event[1] == "download_ipynb":
        (
            _,
            event_args["buf_filepath"],
            event_args["filename"],
        ) = event[2]

        assert event_args["buf_filepath"] != ""
    elif event[1] == "scroll_to_cell":
        (
            _,
            event_args["cursor_pos_row"],
        ) = event[2]
    elif event[1] in ["CursorMoved", "CursorMovedI", "visual_enter", "visual_leave"]:
        (
            _,
            event_args["cursor_pos_row"],
            event_args["visual_start_row"],
        ) = event[2]
    elif event[1] == "grab_entire_buf":
        (
            _,
            event_args["content"],
        ) = event[2]

    event_dict["bufnr"] = event[2][0]
    event_dict["args"] = event_args
    return event_dict


def events_to_listdict(events):
    return [event_to_dict(event) for event in events]
"""



import logging
import time

import pynvim
from pkg_resources import resource_stream

logger = logging.getLogger(__name__)


def attach_and_init(nvim_listen_addr):
    logger.info("nvim addr: %s", nvim_listen_addr)
    for _ in range(30):
        try:
            if ":" in nvim_listen_addr:
                host, port = nvim_listen_addr.split(":")
                nvim = pynvim.attach(
                    "tcp",
                    address=host,
                    port=int(port),
                )
            else:
                nvim = pynvim.attach("socket", path=nvim_listen_addr)
        except Exception:
            time.sleep(0.1)
        else:
            break
    else:
        raise TimeoutError("Timeout while waiting for nvim to start")

    logger.info("nvim attached")

    # existing_channel_id = nvim.vars.get("jupynium_channel_id", None)
    # if existing_channel_id is None:
    logger.info("Initialising..")
    logger.info(f"Communicating with channel_id {nvim.channel_id}")
    nvim.vars["jupynium_channel_id"] = nvim.channel_id
    nvim.vars["jupynium_num_pending_msgs"] = 0
    # Define helper functions
    # Must come at the beginning
    lua_code = resource_stream("jupynium", "lua/defaults.lua").read().decode("utf-8")
    nvim.exec_lua(lua_code)

    lua_code = resource_stream("jupynium", "lua/helpers.lua").read().decode("utf-8")
    nvim.exec_lua(lua_code)

    lua_code = resource_stream("jupynium", "lua/notify.lua").read().decode("utf-8")
    nvim.exec_lua(lua_code)

    # keymaps
    lua_code = resource_stream("jupynium", "lua/commands.lua").read().decode("utf-8")
    nvim.exec_lua(lua_code)
    lua_code = resource_stream("jupynium", "lua/keymaps.lua").read().decode("utf-8")
    nvim.exec_lua(lua_code)

    lua_code = (
        resource_stream("jupynium", "lua/autocmd_vimleave.lua").read().decode("utf-8")
    )
    nvim.exec_lua(lua_code)

    nvim.lua.Jupynium_notify.info(
        [
            "Jupynium successfully attached and initialised.",
            "Run `:JupyniumStartSync`",
        ],
        async_=True,
    )

    return nvim



from __future__ import annotations

import logging

from pkg_resources import resource_stream

from .jupyter_notebook_selenium import insert_cell_at

logger = logging.getLogger(__name__)


set_cell_text_js_code = (
    resource_stream("jupynium", "js/set_cell_text.js").read().decode("utf-8")
)


class JupyniumBuffer:
    """
    This class mainly deals with the Nvim buffer and its cell information.
    This does have a functionality to sync with the Notebook.
    """

    def __init__(self, buf: list[str] = [""]):
        self.buf = buf
        if self.buf == [""]:
            self.num_rows_per_cell: list[int] = [
                1
            ]  # each cell's row length. 0-th cell is not a cell, but it's the header. You can put anything above and it won't be synced to Jupyter Notebook.
            self.cell_types = ["header"]  # 0-th cell is not a cell.
        else:
            self.full_analyse_buf()

    def full_analyse_buf(self):
        num_rows_this_cell = 0
        num_rows_per_cell = []
        cell_types = ["header"]
        for line in self.buf:
            if (
                line.startswith("# %%%")
                or line.startswith("# %% [md]")
                or line.startswith("# %% [markdown]")
                or line.startswith('"""%%')
                or line.startswith("'''%%")
            ):
                num_rows_per_cell.append(num_rows_this_cell)
                num_rows_this_cell = 1
                cell_types.append("markdown")
            elif (
                line.startswith("# %%")
                or line.startswith('%%"""')
                or line.startswith("%%'''")
            ):
                num_rows_per_cell.append(num_rows_this_cell)
                num_rows_this_cell = 1
                cell_types.append("code")
            else:
                num_rows_this_cell += 1
        num_rows_per_cell.append(num_rows_this_cell)

        self.num_rows_per_cell = num_rows_per_cell
        self.cell_types = cell_types

    def process_on_lines(
        self, driver, strip, lines, start_row, old_end_row, new_end_row
    ):
        (
            notebook_cell_operations,
            modified_cell_idx_start,
            modified_cell_idx_end,
        ) = self._on_lines_update_buf(lines, start_row, old_end_row, new_end_row)
        self._apply_cell_operations(driver, notebook_cell_operations)

        num_cells = self.num_cells_in_notebook
        num_cells_in_notebook = driver.execute_script(
            "return Jupyter.notebook.ncells();"
        )

        if num_cells_in_notebook != num_cells:
            self.full_sync_to_notebook(driver, strip=strip)
        else:
            self._partial_sync_to_notebook(
                driver, modified_cell_idx_start, modified_cell_idx_end, strip=strip
            )

    def _on_lines_update_buf(self, lines, start_row, old_end_row, new_end_row):
        """
        Replace start_row:old_end_row to lines from self.buf
        """

        # Analyse how many cells are removed
        notebook_cell_delete_operations = []
        notebook_cell_operations = []

        try:
            cell_idx, cell_start_row, row_within_cell = self.get_cell_index_from_row(
                start_row
            )

            if row_within_cell == 0 and cell_idx > 0:
                # If the row is the first row of a cell, and it's not the first cell, then it's a cell separator.
                row_within_cell = self.num_rows_per_cell[cell_idx - 1]
                cell_idx -= 1
        except IndexError:
            assert start_row == old_end_row == self.num_rows
            cell_idx = self.num_cells - 1
            row_within_cell = self.num_rows_per_cell[-1]

        modified_cell_idx_start = cell_idx

        lines_to_remove = old_end_row - start_row

        while lines_to_remove > 0:
            if row_within_cell < self.num_rows_per_cell[cell_idx]:
                # If the row is within the cell, then it's a cell content.
                self.num_rows_per_cell[cell_idx] -= 1
            else:
                # If the row is not within the cell, then it's a cell separator.
                notebook_cell_delete_operations.append(("delete", cell_idx + 1, None))
                self.num_rows_per_cell[cell_idx] += (
                    self.num_rows_per_cell[cell_idx + 1] - 1
                )
                del self.num_rows_per_cell[cell_idx + 1]
                del self.cell_types[cell_idx + 1]
            lines_to_remove -= 1

        # Analyse how many cells are added
        new_lines_buf = JupyniumBuffer(lines)
        if new_lines_buf.num_cells - 1 == 0:
            self.num_rows_per_cell[cell_idx] += new_lines_buf.num_rows_per_cell[0]
            notebook_cell_operations = notebook_cell_delete_operations
        else:
            num_delete_cells = len(notebook_cell_delete_operations)
            num_insert_cells = new_lines_buf.num_cells - 1
            if num_delete_cells > num_insert_cells:
                notebook_cell_operations = notebook_cell_delete_operations[
                    :num_insert_cells
                ]
            elif num_delete_cells < num_insert_cells:
                notebook_cell_operations = [
                    (
                        "cell_type",
                        cell_idx + 1,
                        new_lines_buf.cell_types[
                            1 : 1 + len(notebook_cell_delete_operations)
                        ],
                    )
                ]
                notebook_cell_operations.append(
                    (
                        "insert",
                        cell_idx + 1,
                        new_lines_buf.cell_types[
                            1 + len(notebook_cell_delete_operations) :
                        ],
                    )
                )
            else:
                notebook_cell_operations = [
                    ("cell_type", cell_idx + 1, new_lines_buf.cell_types[1:])
                ]

            num_tail_rows = self.num_rows_per_cell[cell_idx] - row_within_cell
            self.num_rows_per_cell[cell_idx] = (
                row_within_cell + new_lines_buf.num_rows_per_cell[0]
            )
            new_lines_buf.num_rows_per_cell[-1] += num_tail_rows
            self.num_rows_per_cell[
                cell_idx + 1 : cell_idx + 1
            ] = new_lines_buf.num_rows_per_cell[1:]
            self.cell_types[cell_idx + 1 : cell_idx + 1] = new_lines_buf.cell_types[1:]

        modified_cell_idx_end = modified_cell_idx_start + new_lines_buf.num_cells - 1

        # Now actually replace the lines
        # Optimisation: if the number of lines is not changed, which is most of the cases,
        # then we can just replace the lines.
        if old_end_row == new_end_row:
            for i, line in enumerate(lines):
                self.buf[start_row + i] = line
        else:
            # If the number of lines is changed,
            # then we need to remove the old lines and insert the new lines.
            self.buf[start_row:old_end_row] = lines

        return notebook_cell_operations, modified_cell_idx_start, modified_cell_idx_end

    def _apply_cell_operations(self, driver, notebook_cell_operations):
        # Remove / create cells in Notebook
        for operation, cell_idx, cell_types in notebook_cell_operations:
            nb_cell_idx = cell_idx - 1
            if operation == "delete":
                logger.info(f"Deleting cell {nb_cell_idx} from Notebook")
                driver.execute_script(
                    "Jupyter.notebook.delete_cell(arguments[0]);", nb_cell_idx
                )
            elif operation == "insert":
                for i, cell_type in enumerate(cell_types):
                    logger.info(f"Inserting cell {nb_cell_idx + i} from Notebook")
                    insert_cell_at(driver, cell_type, nb_cell_idx + i)
            elif operation == "cell_type":
                for i, cell_type in enumerate(cell_types):
                    logger.info(
                        f"Cell {nb_cell_idx + i} type change to {cell_type} from Notebook"
                    )
                    if cell_type == "markdown":
                        driver.execute_script(
                            "Jupyter.notebook.cells_to_markdown([arguments[0]]);",
                            nb_cell_idx + i,
                        )
                    elif cell_type == "code":
                        driver.execute_script(
                            "Jupyter.notebook.cells_to_code([arguments[0]]);",
                            nb_cell_idx + i,
                        )
                    else:
                        raise ValueError(f"Unknown cell type {cell_type}")

    def get_cell_start_row(self, cell_idx):
        return sum(self.num_rows_per_cell[:cell_idx])

    def get_cell_index_from_row(self, row):
        """
        Returns the cell index for the given row.

        Returns:
            int: cell index
            int: cell start row
            int: row index within the cell
        """
        cell_start_row = 0
        for i, num_rows in enumerate(self.num_rows_per_cell):
            if cell_start_row + num_rows > row:
                return i, cell_start_row, row - cell_start_row
            cell_start_row += num_rows

        raise IndexError("Could not find cell for row {}".format(row))

    def _check_validity(self):
        assert len(self.buf) == sum(self.num_rows_per_cell)
        assert len(self.cell_types) == len(self.num_rows_per_cell)
        assert self.cell_types[0] == "header"
        assert all(x in ("code", "markdown") for x in self.cell_types[1:])

    def _partial_sync_to_notebook(
        self, driver, start_cell_idx, end_cell_idx, strip=True
    ):
        """
        Cell 1 in JupyniumBuffer is cell 0 in Notebook
        Args are inclusive range in ju.py JupyniumBuffer
        """
        assert start_cell_idx <= end_cell_idx < self.num_cells

        if self.num_cells == 1:
            # Markdown file
            driver.execute_script(
                "Jupyter.notebook.cells_to_markdown([0]);"
                "Jupyter.notebook.get_cell(0).set_text(arguments[0]);"
                "Jupyter.notebook.get_cell(0).render()",
                "\n".join(self.buf),
            )
        else:
            # Notebook file

            if end_cell_idx == 0:
                # Nothing to update
                return

            if start_cell_idx == 0:
                start_cell_idx = 1

            texts_per_cell = []
            start_row = self.get_cell_start_row(start_cell_idx)
            texts_per_cell.append(
                "\n".join(
                    self.buf[
                        start_row
                        + 1 : start_row
                        + self.num_rows_per_cell[start_cell_idx]
                    ]
                )
            )

            for cell_idx in range(start_cell_idx + 1, end_cell_idx + 1):
                start_row += self.num_rows_per_cell[cell_idx - 1]
                texts_per_cell.append(
                    "\n".join(
                        self.buf[
                            start_row + 1 : start_row + self.num_rows_per_cell[cell_idx]
                        ]
                    )
                )

            if strip:
                texts_per_cell = [x.strip() for x in texts_per_cell]

            code_cell_indices = [
                start_cell_idx + i
                for i, cell_type in enumerate(
                    self.cell_types[start_cell_idx : end_cell_idx + 1]
                )
                if cell_type == "code"
            ]
            markdown_cell_indices = [
                start_cell_idx + i
                for i, cell_type in enumerate(
                    self.cell_types[start_cell_idx : end_cell_idx + 1]
                )
                if cell_type == "markdown"
            ]

            if len(code_cell_indices) > 0:
                logger.info(f"Converting to code cells: {code_cell_indices}")
                for i in code_cell_indices:
                    driver.execute_script(
                        "Jupyter.notebook.cells_to_code([arguments[0]]);", i - 1
                    )

            if len(markdown_cell_indices) > 0:
                logger.info(f"Converting to markdown cells: {markdown_cell_indices}")
                for i in markdown_cell_indices:
                    driver.execute_script(
                        "Jupyter.notebook.cells_to_markdown([arguments[0]]);", i - 1
                    )

            # This will render markdown cells
            driver.execute_script(
                set_cell_text_js_code,
                start_cell_idx - 1,
                end_cell_idx - 1,
                *texts_per_cell,
            )

    def full_sync_to_notebook(self, driver, strip=True):
        # Full sync with notebook.
        # WARNING: syncing may result in data loss.
        num_cells = self.num_cells_in_notebook
        num_cells_in_notebook = driver.execute_script(
            "return Jupyter.notebook.ncells();"
        )
        if num_cells > num_cells_in_notebook:
            for _ in range(num_cells - num_cells_in_notebook):
                driver.execute_script("Jupyter.notebook.insert_cell_below();")
        elif num_cells < num_cells_in_notebook:
            for _ in range(num_cells_in_notebook - num_cells):
                driver.execute_script("Jupyter.notebook.delete_cell(-1);")

        self._partial_sync_to_notebook(driver, 0, self.num_cells - 1, strip=strip)

    @property
    def num_cells(self):
        return len(self.num_rows_per_cell)

    @property
    def num_cells_in_notebook(self):
        """
        If the buffer has 1 cell (no separator), it will be treated as markdown file.
        If the buffer has more than 1 cell, it will be treated as notebook.

        Notebook always has 1 cell minimum.
        """
        return max(self.num_cells - 1, 1)

    @property
    def num_rows(self):
        return len(self.buf)

    def __eq__(self, other):
        return (
            self.buf == other.buf
            and self.num_rows_per_cell == other.num_rows_per_cell
            and self.cell_types == other.cell_types
        )



#!/use/bin/env python3
from __future__ import annotations

import argparse
import os

from ..ipynb import ipynb2jupy, load_ipynb


def get_parser():
    parser = argparse.ArgumentParser(
        description="Convert ipynb to a jupynium file (.ju.py)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("ipynb_path", help="Path to ipynb file")
    parser.add_argument(
        "output_jupy_path",
        nargs="?",
        help="Path to output jupynium file. If not specified, use file name of ipynb file or print to stdout (--stdout)",
    )
    parser.add_argument(
        "-y", "--yes", action="store_true", help="Do not ask for confirmation"
    )
    parser.add_argument("-s", "--stdout", action="store_true", help="Print to stdout")
    return parser


def check_args(args, parser):
    if args.stdout and args.yes:
        parser.error("Either one of --stdout or --yes can be specified")

    if args.output_jupy_path is not None and args.stdout:
        parser.error("Either one of --stdout or output_jupy_path can be specified")


def main():
    parser = get_parser()
    args = parser.parse_args()
    check_args(args, parser)

    ipynb = load_ipynb(args.ipynb_path)
    jupy = ipynb2jupy(ipynb)

    if args.stdout:
        for line in jupy:
            print(line)
    else:
        output_jupy_path = args.output_jupy_path
        if output_jupy_path is None:
            output_jupy_path = os.path.splitext(args.ipynb_path)[0] + ".ju.py"

        os.makedirs(os.path.dirname(os.path.realpath(output_jupy_path)), exist_ok=True)

        if os.path.isfile(output_jupy_path) and not args.yes:
            print("Do you want to overwrite {}?".format(output_jupy_path))
            answer = input("y/n: ")
            if answer != "y":
                print("Aborted")
                return

        with open(output_jupy_path, "w") as f:
            for line in jupy:
                f.write(line)
                f.write("\n")

        print('Converted "{}" to "{}"'.format(args.ipynb_path, output_jupy_path))


if __name__ == "__main__":
    main()



# https://stackoverflow.com/questions/36799192/check-if-python-script-is-already-running
from os import getpid
from os.path import exists
from psutil import pid_exists, Process

from .definitions import PACKAGE_DIR


def already_running_pid(name="jupynium"):
    pidpath = PACKAGE_DIR / f"{name}_pid.txt"
    my_pid = getpid()
    if exists(pidpath):
        with open(pidpath) as f:
            pid = f.read()
            pid = int(pid) if pid.isnumeric() else None
        if pid is not None and pid_exists(pid):
            if name in "".join(Process(my_pid).cmdline()) and name in "".join(
                Process(pid).cmdline()
            ):
                return pid
            elif Process(pid).cmdline() == Process(my_pid).cmdline():
                return pid
    with open(pidpath, "w") as f:
        f.write(str(my_pid))
    return 0



import pytest

from jupynium.buffer import JupyniumBuffer


def test_buffer_1():
    buffer = JupyniumBuffer(["a", "b", "c", "# %%", "d", "e", "f"])
    assert buffer.num_rows_per_cell == [3, 4]
    assert buffer.cell_types == ["header", "code"]


def test_buffer_markdown():
    buffer = JupyniumBuffer(["a", "b", "c", "# %%%", "d", "# %%", "f"])
    assert buffer.num_rows_per_cell == [3, 2, 2]
    assert buffer.cell_types == ["header", "markdown", "code"]


def test_buffer_markdown_2(jupbuf1):
    assert jupbuf1.num_rows_per_cell == [3, 2, 2]
    assert jupbuf1.cell_types == ["header", "markdown", "code"]


def test_get_cell_start_row(jupbuf1):
    assert jupbuf1.get_cell_start_row(0) == 0
    assert jupbuf1.get_cell_start_row(1) == 3
    assert jupbuf1.get_cell_start_row(2) == 5


def test_get_cell_index_from_row(jupbuf1):
    assert jupbuf1.get_cell_index_from_row(0) == (0, 0, 0)
    assert jupbuf1.get_cell_index_from_row(1) == (0, 0, 1)
    assert jupbuf1.get_cell_index_from_row(2) == (0, 0, 2)
    assert jupbuf1.get_cell_index_from_row(3) == (1, 3, 0)
    assert jupbuf1.get_cell_index_from_row(4) == (1, 3, 1)
    assert jupbuf1.get_cell_index_from_row(5) == (2, 5, 0)
    assert jupbuf1.get_cell_index_from_row(6) == (2, 5, 1)


def test_check_validity(jupbuf1):
    jupbuf1._check_validity()


@pytest.mark.xfail(raises=Exception)
def test_check_invalid():
    buffer = JupyniumBuffer(["a", "b", "c", "'''%%%", "d", "%%'''", "f"])
    # manually modify the buffer
    buffer.buf.append("g")
    buffer._check_validity()


def test_num_cells(jupbuf1):
    assert jupbuf1.num_cells == 3
    assert jupbuf1.num_cells_in_notebook == 2


def test_num_cells_2():
    buffer = JupyniumBuffer([""])
    assert buffer.num_cells == 1
    assert buffer.num_cells_in_notebook == 1


@pytest.mark.parametrize(
    "content,lines,start_row,old_end_row,new_end_row",
    [
        (["a", "b", "c", "# %%", "d", "e", "f"], ["# %%%", "g"], 3, 4, 5),
        (["b", "c", "# %%", "d", "e", "f"], ["# %%%", "g"], 3, 4, 5),
        (["b", "# %%", "d", "f", "f", "f", "f"], ["# %%"], 3, 4, 4),
        (["b", "# %%", "d", "f", "f", "f", "f"], ["# %%"], 3, 3, 4),
        (["b", "# %%", "d", "# %%", "f", "f", "f"], [""], 1, 4, 2),
        (["b", "# %%", "d", "# %%", "f", "f", "f"], [], 2, 5, 2),
        (["b", "# %%", "d", "# %%", "f", "f", "f"], [], 1, 5, 1),
    ],
)
def test_on_lines_cellinfo(content, lines, start_row, old_end_row, new_end_row):
    buffer = JupyniumBuffer(content)
    # change # %% to # %%% and g
    buffer._on_lines_update_buf(lines, start_row, old_end_row, new_end_row)

    fully_analysed_buf = JupyniumBuffer(buffer.buf)
    assert buffer.num_rows_per_cell == fully_analysed_buf.num_rows_per_cell
    assert buffer.cell_types == fully_analysed_buf.cell_types


@pytest.mark.parametrize(
    "content,lines,start_row,old_end_row,new_end_row,final_content",
    [
        (
            ["a", "b", "c", "# %%", "d", "e", "f"],
            ["# %%%", "g"],
            3,
            4,
            5,
            ["a", "b", "c", "# %%%", "g", "d", "e", "f"],
        ),
        (
            ["b", "c", "# %%", "d", "e", "f"],
            ["# %%%", "g"],
            3,
            4,
            5,
            ["b", "c", "# %%", "# %%%", "g", "e", "f"],
        ),
        (
            ["b", "# %%", "d", "f", "f", "f", "f"],
            ["# %%"],
            3,
            4,
            4,
            ["b", "# %%", "d", "# %%", "f", "f", "f"],
        ),
        (
            ["b", "# %%", "d", "f", "f", "f", "f"],
            ["# %%"],
            3,
            3,
            4,
            ["b", "# %%", "d", "# %%", "f", "f", "f", "f"],
        ),
        (
            ["b", "# %%", "d", "# %%", "f", "f", "f"],
            [],
            1,
            4,
            1,
            ["b", "f", "f", "f"],
        ),
    ],
)
def test_on_lines_content(
    content, lines, start_row, old_end_row, new_end_row, final_content
):
    buffer = JupyniumBuffer(content)
    # change # %% to # %%% and g
    buffer._on_lines_update_buf(lines, start_row, old_end_row, new_end_row)

    final_buf = JupyniumBuffer(final_content)
    assert buffer == final_buf
